package org.dataone.tidy.okay;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import org.dataone.client.CNode;
import org.dataone.client.auth.CertificateManager;
import org.dataone.configuration.Settings;
import org.dataone.service.cn.impl.v1.NodeRegistryService;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.NodeList;
import org.dataone.service.types.v1.util.NodelistUtil;
import org.dataone.tidy.okay.metacat.MetacatDAO;
import org.dataone.tidy.okay.metacat.MetacatProperties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.core.util.StatusPrinter;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import org.dataone.client.D1Client;
import org.dataone.service.exceptions.InvalidRequest;
import org.dataone.service.exceptions.InvalidToken;
import org.dataone.service.exceptions.NotAuthorized;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.NotImplemented;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.v1.Node;
import org.dataone.service.types.v1.ObjectInfo;
import org.dataone.service.types.v1.ObjectList;
import org.dataone.service.types.v1.Session;

/**
 * Hello world!
 *
 */
public class Main {

    private static final Logger logger = LoggerFactory.getLogger(Main.class);
    static Properties metacatProperties = null;
    static final String localNodeId = Settings.getConfiguration().getString("cn.nodeId");
    static private int start = 0;
    static private int total = 0;
    static Integer batchSize = new Integer(10000);

    static {

        String clientCertLocation = Settings.getConfiguration()
                .getString("D1Client.certificate.directory");
        String clientCertFilename = Settings.getConfiguration()
                .getString("D1Client.certificate.filename");
        File d1CertFile = new File(clientCertLocation + File.separator + clientCertFilename);
        String clientCertificateLocation = d1CertFile.getAbsolutePath();
        CertificateManager.getInstance().setCertificateLocation(clientCertificateLocation);

    }

    public static void main(String[] args) {

        try {
            metacatProperties = MetacatProperties.getMetacatProperties();
        } catch (Exception ex) {
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
            System.exit(1);
        }
        LoggerContext lc = (LoggerContext) LoggerFactory.getILoggerFactory();
        StatusPrinter.print(lc);

        NodeRegistryService nodeRegistry = new NodeRegistryService();
        NodeList nodeList = new NodeList();
        try {
            nodeList = nodeRegistry.listNodes();
        } catch (NotImplemented ex) {
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
            System.exit(1);
        } catch (ServiceFailure ex) {
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
            System.exit(1);
        }
        Map<String, String> baseUrlMap = NodelistUtil.mapNodeList(nodeList);
        String localBaseURL = baseUrlMap.get(localNodeId);
        Settings.getConfiguration().setProperty("D1Client.CN_URL", localBaseURL);
        D1Client d1Client = new D1Client();
        CNode cnNode = null;
        try {
            cnNode = d1Client.getCN();
        } catch (ServiceFailure ex) {
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
            System.exit(1);
        }
        MetacatDAO metacatDao = new MetacatDAO();
        boolean okay = true;
        Set<String> metacatIdentifierSet = null;
        try {
            metacatIdentifierSet = metacatDao.getMetacatIdentifiers();
        } catch (SQLException ex) {
            okay = false;
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
        } catch (IOException ex) {
            okay = false;
            logger.error(ex.getLocalizedMessage());
            ex.printStackTrace();
        }
        if (okay) {
            List<String> idlist = null;
            try {
                idlist = listIdentifiers(cnNode);
            } catch (InvalidRequest ex) {
                okay = false;
                logger.error(ex.getLocalizedMessage());
                ex.printStackTrace();
            } catch (InvalidToken ex) {
                okay = false;
                logger.error(ex.getLocalizedMessage());
                ex.printStackTrace();
            } catch (NotAuthorized ex) {
                okay = false;
                logger.error(ex.getLocalizedMessage());
                ex.printStackTrace();
            } catch (NotImplemented ex) {
                okay = false;
                logger.error(ex.getLocalizedMessage());
                ex.printStackTrace();
            } catch (ServiceFailure ex) {
                okay = false;
                logger.error(ex.getLocalizedMessage());
                ex.printStackTrace();
            }
            if (okay) {
                for (String pid : idlist) {
                    if (!metacatIdentifierSet.contains(pid)) {
                        okay = false;
                        logger.warn(pid + " was found in database but not in list objects");
                    }
                }
            }
            if (okay) {
                logger.info("listObjects appears okay with " + idlist.size() + " # of objects from a count of " + total);
            } else {
                logger.warn("listOjects failed");
            }
        }

        okay = true;
        for (Node mnNode : nodeList.getNodeList()) {
            if (mnNode.getType().equals(org.dataone.service.types.v1.NodeType.MN)) {
                try {
                    metacatIdentifierSet = metacatDao.getMetacatIdentifiersByAuthNode(mnNode.getIdentifier().getValue());
                } catch (SQLException ex) {
                    okay = false;
                    logger.error(ex.getLocalizedMessage());
                    ex.printStackTrace();
                } catch (IOException ex) {
                    okay = false;
                    logger.error(ex.getLocalizedMessage());
                    ex.printStackTrace();
                }
                if (okay) {
                    for (String pidStr : metacatIdentifierSet) {
                        Identifier pid = new Identifier();
                        pid.setValue(pidStr);
                        try {
                            cnNode.getSystemMetadata(pid);
                        } catch (InvalidToken ex) {
                            okay = false;
                            logger.error(ex.getLocalizedMessage());
                            ex.printStackTrace();
                        } catch (ServiceFailure ex) {
                            okay = false;
                            logger.error(ex.getLocalizedMessage());
                            ex.printStackTrace();
                        } catch (NotAuthorized ex) {
                            okay = false;
                            logger.error(ex.getLocalizedMessage());
                            ex.printStackTrace();
                        } catch (NotFound ex) {
                            okay = false;
                            logger.error(ex.getLocalizedMessage());
                            ex.printStackTrace();
                        } catch (NotImplemented ex) {
                            okay = false;
                            logger.error(ex.getLocalizedMessage());
                            ex.printStackTrace();
                        }
                    }
                }
            }
        }
        if (okay) {
            logger.info("getSystemMetadatas appear okay");
            logger.info("d1 tidy is okay");
        } else {
            logger.warn("getSystemMetadatas failed");
        }
        System.exit(0);
    }

    private static List<String> listIdentifiers(CNode cnNode) throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure {
        List<String> listIds = new ArrayList<String>();

        List<ObjectInfo> readQueue = null;

        do {
            if (start == 0 || (start < total)) {
                readQueue = retrieve(cnNode);
                for (ObjectInfo objectInfo : readQueue) {
                    listIds.add(objectInfo.getIdentifier().getValue());

                }
            } else {
                readQueue = null;
            }
        } while ((readQueue != null) && (!readQueue.isEmpty()));

        return listIds;
    }

    private static List<ObjectInfo> retrieve(CNode cnNode) throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure {
        // logger is not  be serializable, but no need to make it transient imo

        List<ObjectInfo> writeQueue = new ArrayList<ObjectInfo>();

        ObjectList objectList = null;
        Boolean replicationStatus = null;
        Session session = new Session();

            // always execute for the first run (for start = 0)
        // otherwise skip because when the start is equal or greater
        // then total, then all objects have been harvested
        objectList = cnNode.listObjects(session, null, null, null, null, start, batchSize);
            // if objectList is null or the count is 0 or the list is empty, then
        // there is nothing to process
        if (!((objectList == null)
                || (objectList.getCount() == 0)
                || (objectList.getObjectInfoList().isEmpty()))) {

            start += objectList.getCount();
            writeQueue.addAll(objectList.getObjectInfoList());
            total = objectList.getTotal();

        }

        return writeQueue;
    }
}
